// ===========================================================================
//
//	IUGSTAS.cpp   			 Copyright R.W. Le Maitre 2002 - Version 1.0
//
// ===========================================================================
//  A class for reading in rock analyses, calculating CIPW norms and classifying
//	them using the IUGS TAS classification.

#include "IUGSTAS.h"

#define 	Minimum(x,y) ( ((x) < (y)) ? (x) : (y) )

const	double 	kVerySmall	= 1e-10;

// 
//		 IUGSTAS						(constructor)
// 
IUGSTAS::IUGSTAS()
{
	int i;
	
	inputDefined = false;
	inputRecRead = 0;
	
	for (i = 0; i < numOutputFiles; i++) {
		outputDefined[i] = false;
		outputRecWritten[i] = 0;
	}
	
	multiColTableInUse = -1;
		
	Init();
}

// 
//		 ~IUGSTAS						(destructor)
// 
IUGSTAS::~IUGSTAS()
{
	int i;
	
	if(inputDefined) inputFile.close();
	for(i = 0; i < numOutputFiles; i++) {
		if(outputDefined[i]) {
			outputFile[i].close();
		}
	}
}

// 
//		 Init		(public)
// 
void 
IUGSTAS::Init()
{
	int i;

	specimen.name[0] = '\0';									
	TASFieldSymbol[0] = '\0';									
	TASOptName[0] = '\0';									
	TASRootName[0] = '\0';		

	for(i = 0; i < numOxides; i++) oxide.wt[i] = 0.0;
	for(i = 0; i < numNormValues; i++) norm.wt[i] = 0.0; 	
}

#pragma mark -

// 
//		 AskForInputFile		(public)
// 
bool
IUGSTAS::AskForInputFile(char* taskName)
{
	bool 		openOK;
	char		fName[lenFileName];
	
	do {
		cout << "Enter name of tab delimited file to be read or RETURN to exit task";
		if(taskName[0] != '\0') cout << " " << taskName;
		cout << ":  " << endl;
		
		cin.getline(fName, lenFileName);
		cout << endl;
		
		if(strlen(fName) == 0)  return false;

		openOK = OpenInputFile(fName);

	} while (!openOK);
	return true;
}

// 
//		 OpenInputFile		(public)
// 
bool
IUGSTAS::OpenInputFile(char *fName)
{
	inputFile.open( fName, ios::in);
	bool openOK = !inputFile.fail();
	
	if(openOK) {
		inputRow = 0;
		inputCol = 0;
		inputDefined = true;
		inputHeaderRead = false;
		strcpy((char*)&inputFileName, fName);
	} else {
		Warning ("File \"", fName, "\" cannot be opened for input. ",
			"It may not exist or may already be open.");
		inputFile.clear();
	}
	return openOK;
}

// 
//		 AskForOutputTables		(public)
// 
bool
IUGSTAS::AskForOutputTables(char* taskName)
{
	char			fName[lenFileName];
	unsigned long	numTables, i;
	
	do {
		cout << endl;
		cout << "Enter up to 3 characters for types of tables to be written" << endl;
		cout << "   S for a single half-page table" << endl;
		cout << "   V for a table with analyses in vertical columns" << endl;
		cout << "   H for a table with analyses in horizontal rows" << endl;
		cout << " 	or RETURN to exit task";
		if(taskName[0] != '\0') cout << " " << taskName;
		cout << ":  ";
		 
		cin.getline(textInput, lenTextInput);
		numTables = strlen(textInput);

		for (i = 0; i < numOutputFiles; i++) {
			textInput[i] = (char)toupper(textInput[i]);
			if(i < numTables) {
				outputTableType[i] = textInput[i];
			} else {
				outputTableType[i] = ' ';
			}
		}
		
		if(numTables > 3) {
			Warning("no more than three characters may be entered.");
		} else if(numTables == 0) {
			return false;
		} else if (strchr(textInput, 'S') != strrchr(textInput,'S')) {
			Warning("the character S may only occur once.");
		} else if (strchr(textInput, 'V') != strrchr(textInput,'V')) {
			Warning("the character V may only occur once.");
		} else if (strchr(textInput, 'H') != strrchr(textInput,'H')) {
			Warning("the character H may only occur once.");
		} else if(!strchr("SVH", textInput[0]) ||
					(numTables > 1 && !strchr("SVH", textInput[1])) ||
					(numTables > 2 && !strchr("SVH", textInput[2]))) {
			Warning("Only the characters S V and H may be entered.");
		} else {			
			for (i = 0; i < numTables; i++) {
				outputTableType[i] = (char)toupper(textInput[i]);
				CreateAutoFileName((char*)&inputFileName, fName, (char)toupper(textInput[i]));
				cout << "Output file is \"" << fName << "\". RETURN if OK, otherwise type new name:  ";
				cin.getline(textInput, lenTextInput);
				if(textInput[0] != '\0')  {
					strcpy(fName, textInput);
				}
				if(!OpenOutputFile(fName, (int)(i + 1))) {
					CloseAllOutputFiles();
					return false;
				}
			}
			return true;
		}		
	}while(true);
	
	return true;
}

// 
//		 OpenOutputFile		(public)
// 
bool
IUGSTAS::OpenOutputFile(char *fName, int nos)
{
	CheckOutputFile(fName, nos, "OpenOutputFile");	

	outputFile[nos].open(fName, ios::out);
	if(outputFile[nos].fail()) {
		Warning ("File \"", fName, "\" cannot be opened for output. ",
			"It may already be open.");
		outputFile[nos].clear();
		return false;
	}
	columnsStored = 0;
	tableSeq = 1;
	outputDefined[nos] = true;
	outputHeaderWritten[nos] = false;
	strcpy((char*)&outputFileName[nos], fName);
	return true;
}

// 
//		 CloseInputFile		(public)
// 
void
IUGSTAS::CloseInputFile()
{
	inputFile.clear();
	inputFile.close();
	
	inputDefined = false;
	inputHeaderRead = false;
	inputRecRead = 0;
	strcpy((char*)&inputFileName, "");
}

// 
//		 CloseOutputFile		(public)
// 
void
IUGSTAS::CloseOutputFile(int nos)
{
	CheckOutputFile("", nos, "CloseOutputFile");
	
	outputFile[nos].close();
	outputFile[nos].clear();
		
	outputDefined[nos] = false;
	outputHeaderWritten[nos] = false;
	outputRecWritten[nos] = 0;
	outputTableType[nos] = ' ';
	strcpy((char*)&outputFileName[nos], "");
	multiColTableInUse = -1;
}

// 
//		 CloseAllOutputFiles		(public)
// 
void
IUGSTAS::CloseAllOutputFiles()
{
	for(int i = 1; i <= numOutputFiles; i++) {
		CloseOutputFile(i);
	}
}

#pragma mark -

// 
//		 ReadTabDelimitedRow		(public)
// 
void
IUGSTAS::ReadTabDelimitedRow()
{
	if(!inputDefined) {
		AbortProgramingError("ReadTabDelimitedRow", 
					"This routine can only be used after a matching call to",
					"either AskForInputFile() or OpenInputFile() has been made.");
	}

	if (!inputHeaderRead) {
		ReadHeader(inputFile);
		inputHeaderRead = true;
	}
	
	do {
		validRow = false;
		GetTabDelimitedStr(inputFile, specimen.name, lenSpecName);
		if(specimen.name[0] > 0 ) validRow = true;
		
		if(lastReadResult == kOK) {
		
			int i = 0;
			inputCol = 1;
			validRow = false;
			
			do{
				if(i < numOxides) {
					double 	val;					
					GetDouble(inputFile, val);
					oxide.wt[oxOrder[i++]] = val;
					if(val > 0 ) validRow = true;
				} else {
					char	temp[64];
					GetTabDelimitedStr(inputFile, temp, 64);
				}
				if(lastReadResult == kDataTruncated) return;
				
				inputCol++;
				
			} while (lastReadResult != kEndOfLine && lastReadResult != kEndOfFile);
		}
		inputRow++;

	} while (!validRow && lastReadResult != kEndOfFile);
	
	if(validRow) inputRecRead++;
}

// 
//		 ReadHeader		(protected)
// 
void
IUGSTAS::ReadHeader(ifstream &input)
{
	char	temp[lenSpecName];
	int		i;
	
	for(i = 0; i < numOxides; i++) {
		oxOrder[i] = -1;
		oxUsed[i] = false;
	}

	input.seekg(0);									// rewind file
	GetTabDelimitedStr(input, temp, lenSpecName);	// Ignore analysis name
	inputRow = 1;
	inputCol = 1;
	
	do{
		GetTabDelimitedStr(input, temp, lenSpecName);
		if(inputCol == 1 && IsPositiveNumber(temp)) {	// assume it is data
			for(i = 0; i < numOxides; i++) {
				oxOrder[i] = i;
			}
			input.seekg(0);
			return;
		}
		
		int nth = IsValidOxide(temp);
		
		if(inputCol < numOxides && nth == -1) {
			Warning("unrecognized oxide named \"", temp, "\" will be ignored.");
		}
		
		oxOrder[inputCol - 1] = nth;
		
		if(nth > -1) {
			if(oxUsed[nth]) {
				AbortDataError(temp, "Oxide name has already been used.");
			} else {
				oxUsed[nth] = true;
			}
		}
		
		inputCol++;
		
	} while (lastReadResult != kEndOfLine && lastReadResult != kEndOfFile);
}

// 
//		 EndOfFile		(public)
// 
bool
IUGSTAS::EndOfFile()
{
	bool finished = (lastReadResult == kEndOfFile) && GetAnalysisTotal() < 0.1;
	
	if(inputRecRead == 1) {
		cout << endl << "Starting reading." << endl;
	}

	if(finished) {
		if(multiColTableInUse > -1) {
			WriteVertTable(outputFile[multiColTableInUse]);
		}	
		cout << endl << endl;
		if(inputDefined) {
			cout << inputRecRead << " records read from file \"" 
							<< (char *)&inputFileName << "\"." << endl;
		}			
		for (int i = 0; i < numOutputFiles; i++) {
			if(outputDefined[i]) {
				cout << outputRecWritten[i] << " records written to file \"" 
							<< (char *)&outputFileName[i] << "\"." << endl;
			}			
		}
		cout << endl;
	} else {
		if(fmod((double)inputRecRead, reportEveryNth) == 0) {
			cout << inputRecRead << " records read." << endl;
		}
	}
	
	return finished;
}

// 
//		 GetTabDelimitedStr		(protected)
// 
void
IUGSTAS::GetTabDelimitedStr(ifstream &input, char *inStr, int maxLen)
{
	char		ch, chtc;
	int			i = 0;
	
	do {
		input.get(ch);
		chtc = '\0';
		
		if(input.eof()) {
			*(inStr + i) = '\0';
			lastReadResult = kEndOfFile;
			return;
		 }
		 
		switch (ch) {
			case '\t':
				lastReadResult = kOK;
				break;				
			case '\n':
				lastReadResult = kEndOfLine;
				break;				
			case '\r':
				input.get(ch);
				if(ch != '\n') {
					*(inStr + i) = '\0';
					AbortDataError(inStr, "String terminated with Carriage Return not followed by Line Feed.");
				}
				lastReadResult = kEndOfLine;
				break;				
			default:
				chtc = ch;
				break;	
		}
		*(inStr + i) = chtc;
		i++;
	} while (i < maxLen && chtc != '\0');
	
	if(chtc == '\0') return;
	*(inStr + i) = '\0';
	lastReadResult = kDataTruncated;
	return;
}

// 
//		 GetDouble		(protected)
// 
void
IUGSTAS::GetDouble(ifstream &input, double &f)
{
	char 		aStr[16];
	
	GetTabDelimitedStr(input, aStr, 16);
	if(lastReadResult == kDataTruncated) return;
	
	if(strlen(aStr) == 0) {
		f = 0.0;
	} else {
		if(IsPositiveNumber(aStr)) {
			istrstream ii(aStr, 16);
			ii >> f;
		} else {
			AbortDataError(aStr, "Oxide values must be positive numbers.");
		}
	}
	return;
}

// 
//		 GetOxidesInterActively		(public)
// 
void 
IUGSTAS::GetOxidesInterActively()
{
	bool		numberOK;
	int			i = 0, j;	

	cout << endl << "For each oxide displayed enter:";
	cout << endl << "    a NEW VALUE or";
	cout << endl << "    RETURN to leave current value unchanged or ";
	cout << endl << "    any non-numeric character if finished" << endl << endl;
	
	for(j = 0; j < numNormValues; j++) norm.wt[j] = 0.0; 	
	
	do {
		do {
			numberOK = true;
			cout << (char*)&oNames[i] <<  " = "
				 << setiosflags(ios::fixed) << setprecision(2) << oxide.wt[i] << ":  ";
				
			cin.getline(textInput, lenTextInput);
			if(textInput[0] != '\0') {
				if(!isdigit(textInput[0]) && textInput[1] == '\0') {
					for(j = i; j < numOxides - 3; j++) oxide.wt[j] = 0.0; 
					return;
				} else {
					numberOK = IsPositiveNumber(textInput);
					if(numberOK) {
						istrstream ii(textInput, lenTextInput);
						ii >> oxide.wt[i];
					} 
				}
			}
		} while(!numberOK);
		i++;
		
	} while (i < numOxides - 3);    
}

// 
//		 GetWaterCO2		(public)
// 
bool 
IUGSTAS::GetWaterCO2(ERecalculate &recalcMode, ECO2Usage &CO2usage, char* taskName)
{
	cout << "Enter 1 to recalculate to 100% IGNORING water and CO2" << endl;
	cout << "             (THE SUBCOMMISSION'S RECOMMENDED OPTION)" << endl;
	cout << "      2 to recalculate to 100% WITH water and CO2" << endl;
	cout << "      3 DO NOT recalculate the analysis" << endl;
	cout << "      any other character to exit task";
	if(taskName[0] != '\0') cout << " " << taskName;
	cout << ":  ";

	cin.getline(textInput, lenTextInput);
	cout << endl;
	
	switch (textInput[0]) {
		case '1':
			recalcMode = kWaterCO2Free;
			break;
		
		case '2':
			recalcMode = kWithWaterCO2;
			break;
	
		case '3':
			recalcMode = kDontRecalculate;
			break;
	
		default:
			return false;
	}

	if(recalcMode != kWithWaterCO2) return true;
	
	cout << "Enter 1 to calculate CO2 as Na2CO3" << endl;
	cout << "      2 to calculate CO2 as CaCO3, MgCO3 and FeCO3" << endl;
	cout << "      3 to put CO2 in others  " << endl;
	cout << "      any other character to exit task";
	if(taskName[0] != '\0') cout << " " << taskName;
	cout << ":  ";

	cin.getline(textInput, lenTextInput);
	cout << endl;

	switch (textInput[0]) {
		case '1':
			CO2usage = kCO2AsNaCarb;
			break;
		
		case '2':
			CO2usage = kCO2AsCaMgFeCarb;
			break;
	
		case '3':
			CO2usage = kCO2AsOthers;
			break;
	
		default:
			return false;
	}
	return true;
}

// 
//		 IsValidOxide		(public)
// 
int
IUGSTAS::IsValidOxide(char *inStr)
{
	int 		i;
	
	for(i = 0; i < numOxides - 3; i++) {
		if(StringsEqual(inStr, (char *)&oNames[i])) return i;
	}
	return -1;
}

// 
//		 StringsEqual		(public)
// 
bool
IUGSTAS::StringsEqual(char *inStr1, char * inStr2)
{
	int i = 0;
	
	while (toupper(inStr1[i]) == toupper(inStr2[i]) && inStr1[i] != '\0' && inStr2[i] != '\0' ) {
		i++;
		if(inStr1[i] == '\0' && inStr2[i] == '\0') return true;
	}
	return false;
}

#pragma mark -

// 
//		 WriteTables		(public)
// 
void
IUGSTAS::WriteTables()
{
	for (int i = 1; i <= numOutputFiles; i++) {
		switch(outputTableType[i - 1]) {
			case 'S':
				WriteTable(i);
				break;
			case 'H':
				WriteAsHorzTable(i);
				break;
			case 'V':
				WriteAsVertTable(i, numTableCols, false);
				break;
		}	
	}
}

// 
//		 WriteAsHorzTable		(public)
// 
void
IUGSTAS::WriteAsHorzTable(int nos)
{
	if(validRow) {
		CheckOutputFile("", nos, "WriteAsHorzTable");	
		if (!outputHeaderWritten[nos]) {
			WriteHeader(outputFile[nos]);
			outputHeaderWritten[nos] = true;
		}
		outputRecWritten[nos]++;
		WriteAsHorzTable(outputFile[nos]);
	}
}

// 
//		 WriteAsHorzTable		(public)
// 
void
IUGSTAS::WriteAsHorzTable(ostream &fout)
{
	if(validRow) {
		int i;

		fout << specimen.name;
		fout << setiosflags(ios::fixed) << setprecision(2);

		for (i = 0; i < numOxides; i++ ) fout << tabChar << oxide.wt[i];
		for (i = 0; i < numNormValues; i++ ) fout << tabChar << norm.wt[i];

		fout << endl;
		
		if(fout.fail()) {
			UnexpectedWriteError("WriteAsHorzTable()", specimen.name);
			fout.clear();
		}
	}
}

// 
//		 WriteHeader		(public)
// 
void
IUGSTAS::WriteHeader(ostream &fout)
{
	int i;

	fout << "Specimen Name";	
	for (i = 0; i < numOxides; i++ ) fout << tabChar << (char*)&oNames[i];
	for (i = 0; i < numNormValues; i++ ) fout << tabChar << (char*)&nNames[i];
	
	fout << endl;
}

// 
//		 WriteTable		(public)
// 
void
IUGSTAS::WriteTable(int nos)
{
	if(validRow) {
		CheckOutputFile("", nos, "WriteTable");
		
		outputRecWritten[nos]++;
		WriteTable(outputFile[nos]);
	}	
}

// 
//		 WriteTable		(public)
// 
void
IUGSTAS::WriteTable(ostream &fout)
{
	if(validRow) {
		int	col1 = 15, col3 = 16, col4 = 32;
		int	n2 = col1, n4 = col3, n5 = col4;
	 
	 	fout << specimen.name << endl << setiosflags(ios::fixed) << setprecision(2);
		fout << endl; 

		for (int n = 0; n < 16; n++) {
			PlaceStrAndValue(fout, 0, (char*)&oNames[n], 6, oxide.wt[n], 6 , n < col1);
			PlaceStrAndValue(fout, 2, (char*)&oNames[n2], 6, oxide.wt[n2], 6, n2 < numOxides && n != 13);
			if(n != 13) n2++;
			PlaceStrAndValue(fout, 6, (char*)&nNames[n], 4, norm.wt[n], 6 , n < col3);
			PlaceStrAndValue(fout, 2, (char*)&nNames[n4], 4, norm.wt[n4], 6, n4 < col4);
			n4++;
			PlaceStrAndValue(fout, 2, (char*)&nNames[n5], 7, norm.wt[n5], 6, n5 < numNormValues && n != 13);
			if(n != 13) n5++;
			fout << endl;
		}
		fout << endl;

		if(fout.fail()) {
			UnexpectedWriteError("WriteTable()", specimen.name);
			fout.clear();
		}
	}
}

// 
//		 WriteAsVertTable		(public)
// 
void
IUGSTAS::WriteAsVertTable(int nos, int numCols, bool printAll)
{
	if(validRow) {
		CheckOutputFile("", nos, "WriteTablulated");

		if(numCols > numTableCols) AbortProgramingError("WriteAsVertTable", 
					"Maximum number of columns printed must be between 1 and ", numTableCols, ""); 
		if(multiColTableInUse == -1) multiColTableInUse = nos;
		if(multiColTableInUse != nos) AbortProgramingError("WriteAsVertTable", 
											"This routine can only be used once in a task.", ""); 
		outputRecWritten[nos]++;

		WriteAsVertTable(outputFile[nos], numCols, printAll);
	}	
}

// 
//		 WriteAsVertTable		(public)
// 
void
IUGSTAS::WriteAsVertTable(ofstream &fout, int numCols, bool printAll)
{
	if(validRow) {
		data[columnsStored].oxide = oxide;
		data[columnsStored].norm = norm;
		strcpy(data[columnsStored].specimen.name, specimen.name);
		columnsStored++;
		printAllRows = printAll;
		
		if(columnsStored >= numTableCols || columnsStored >= numCols) 
												WriteVertTable(fout);	
	}
}

// 
//		 WriteVertTable		(protected)
// 
void
IUGSTAS::WriteVertTable(ofstream &fout)
{
	int 	i, nthCol;

 	fout << setiosflags(ios::fixed) << setprecision(2);
 	
	for (nthCol = 0; nthCol < columnsStored; nthCol++)  {
		fout << tabChar<< tableSeq;
		tableSeq++;
	}
	
	fout << endl << endl;

	for (i = 0; i < numOxides; i ++ ) {
		if(printAllRows || RowHasValues(i, true)) {
			fout << (char*)&oNames[i];
			for (nthCol = 0; nthCol < columnsStored; nthCol++) {
				fout << tabChar << data[nthCol].oxide.wt[i];
			}
			fout << endl;
		}
	}

	fout << endl << "CIPW Norm" << endl;

	for (i = 0; i < numNormValues; i ++ ) {
		if(printAllRows || RowHasValues(i, false)) {
			fout << (char*)&nNames[i];
			for (nthCol = 0; nthCol < columnsStored; nthCol++) {
				fout << tabChar << data[nthCol].norm.wt[i];
			}
			fout << endl;
		}
	}

	fout << endl << "Specimen Names" << endl;
	
	tableSeq -= columnsStored;

	for (nthCol = 0; nthCol < columnsStored; nthCol++)  {
		fout << tableSeq << ") " << data[nthCol].specimen.name << endl;
		tableSeq++;
	}
	fout << endl << endl;
	columnsStored = 0;

	if(fout.fail()) {
		UnexpectedWriteError("WriteAsVertTable()", data[0].specimen.name);
		fout.clear();
	}
}

// 
//		 RowHasValues		(protected)
// 
bool
IUGSTAS::RowHasValues(int i, bool isOxide)
{
	for (int nthCol = 0; nthCol < columnsStored; nthCol++) {
		if(isOxide) {
			if(data[nthCol].oxide.wt[i] != 0.0)  return true;
		} else {
			if(data[nthCol].norm.wt[i] != 0.0)  return true;
		}
	}
	return false;
}

// 
//		 AmendSpecimenName		(protected)
// 
void
IUGSTAS::AmendSpecimenName()
{	if(TASRootName[0] > 0) {
		char temp[lenSpecName];
		bool truncation = false;		
		ostrstream ii(temp, lenSpecName);
		
	 	ii <<  " = " << TASRootName;
	 	if(TASFieldSymbol[0] > 0) ii << " (" << TASFieldSymbol << ")";
	 	if(TASOptName[0] > 0) ii << " [" << TASOptName << "]"; 
	 	ii << '\0';
	 	
	 	int nLeft =  lenSpecName - (int)strlen(specimen.name) - 1;
	 	if( nLeft < strlen(temp)) {
	 		temp[nLeft-1] = '';
	 		temp[nLeft] = '\0';
	 		truncation = true;
	 	} 
	 	strcat(specimen.name, temp);
	 	if(truncation) {
			cout << endl << endl << sixAst;
			cout << "Warning - specimem name with appended TAS names truncated:" << endl 
					<< specimen.name << endl << endl;
		}
 	}
}

// 
//		 PlaceStrAndValue		(protected)
// 
//	if printIt is true it prints inStr and inValue, otherwise it prints spaces only
void 
IUGSTAS::PlaceStrAndValue(ostream &fout, int initSpaces, char *inStr, int widStr, 
								double inValue, int widVal, bool printIt)
{
	if(printIt) {	
		if(initSpaces > 0) fout << setw(initSpaces) << " ";
		fout.setf(ios::left, ios::adjustfield);
		fout << setw(widStr) << inStr;
		fout.setf(ios::right, ios::adjustfield);
		fout << setw(widVal) << inValue;
	} else {
		fout << setw(initSpaces + widStr + widVal) << " ";
	}
}

// 
//		 CreateAutoFileName		(protected)
// 
void
IUGSTAS::CreateAutoFileName(char* oldName, char* newName, char tableType)
{
	char		appendStr[3] = "_ ";
	char		*ptr = strchr(oldName, '.');
	
	appendStr[1] = tableType;
	
	if(ptr) {
		size_t	lStr = (unsigned long)(ptr - oldName);
		strncpy(newName, oldName, lStr);
		newName[lStr] = '\0';
		strcat(newName, (char*)&appendStr);
		strcat(newName, (char*)&oldName[lStr]);
	} else {
		strcpy(newName, oldName);
		strcat(newName, (char*)&appendStr);	
	}
}

#pragma mark -
// 
//		 GetTASName		(public)
// 
void
IUGSTAS::GetTASName(ERecalculate recalcMode, ECO2Usage CO2usage)
{
	GetTASName(TASOptName, TASRootName, TASFieldSymbol, oxide, norm, recalcMode, CO2usage);
	AmendSpecimenName();
}

// 
//		 GetTASName		(public)
// 
void
IUGSTAS::GetTASName(char *optName, char *rootName, char *fieldSymbol, SOxidepc &ox, 
								SNormpc &norm, ERecalculate recalcMode, ECO2Usage CO2usage)			
{
	SLine	TASLine[10] = 
		{{ 0.176471, -4.176471},	// Line 0 from (52.0, 5.0) to (69.0, 8.0) 
		{-1.000000, 77.000000},		// Line 1 from (69.0, 8.0) to (77.0, 0.0) 
		{0.531746, -18.928571},		// Line 2 from (45.0, 5.0) to (57.6, 11.7) 
		{-0.870370, 61.833317},		// Line 3 from (57.6, 11.7) to (63.0, 7.0) 
		{-0.850000, 54.350002},		// Line 4 from (53.0, 9.3) to (57.0, 5.9) 
		{-0.884616, 51.000031},		// Line 5 from (49.4, 7.3) to (52.0, 5.0) 
		{0.608696, -17.956520},		// Line 6 from (41.0, 7.0) to (52.5, 14.0) 
		{-0.450981, 37.676479},		// Line 7 from (57.6, 11.7) to (52.5, 14.0) 
		{-0.478261, 34.647835},		// Line 8 from (53.0, 9.3) to (48.4, 11.5) 
		{-0.477272, 30.877258}};	// Line 9 from (49.4, 7.3) to (45.0, 9.4)

	Recalculate(ox, recalcMode);
	
	CalcCIPW(CO2usage, ox, norm);
	
	optName[0] = '\0';
	rootName[0] = '\0';
	fieldSymbol[0] = '\0';

	if(oxide.named.SiO2 > 52.0 && oxide.named.MgO > 8.0 && oxide.named.TiO2 < 0.5) {	// High-Mg	
		strcpy(rootName, "Boninite");
		return;
	} else if (oxide.named.SiO2 > 30.0 && oxide.named.SiO2 < 52.0 && oxide.named.MgO > 12.0 && oxide.named.TAlk < 3.0) {
		if(oxide.named.MgO > 18 && oxide.named.TAlk < 2.0) {
			if(oxide.named.TiO2 < 1.0) {
				strcpy(rootName, "Komatiite");			
			} else {
				strcpy(rootName, "Meimechite");
			}
		} else {	
			strcpy(rootName, "Picrite");
		}
		return;
	} 
	
	if (oxide.named.SiO2 <= 52.0 && oxide.named.SiO2 > 45.0 && oxide.named.TAlk <= 5.0) {	// B
		strcpy(fieldSymbol, "B");
		strcpy(rootName, "Basalt");
		if(norm.named.ne > 0.0 ) {
			strcpy(optName, "Alkali");
		} else{
			strcpy(optName, "Subalkali");
		}
		return;
	}
	
	if(oxide.named.SiO2 > 69.0 && AboveLine(TASLine[1])) {	// R
		strcpy(fieldSymbol, "R");
		strcpy(rootName, "Rhyolite");
		if(IsPeralkaline(ox)) {
			IsComenditic(optName, ox);	
		} else {
			GetKState(optName, ox);
		}
		return;
	}
	
	if(oxide.named.SiO2 > 52.0 && !AboveLine(TASLine[0])) { // O1 - O3
		GetKState(optName, ox);
		if(oxide.named.SiO2 <= 57.0) {
			strcpy(fieldSymbol, "O1");
			strcpy(rootName, "Basaltic Andesite");
		} else if(oxide.named.SiO2 <= 63.0) {
			strcpy(fieldSymbol, "O2");
			strcpy(rootName, "Andesite");
		} else {
			strcpy(fieldSymbol, "O3");
			strcpy(rootName, "Dacite");
		}
		return;
	}

	if(oxide.named.SiO2 > 45.0 && !AboveLine(TASLine[2])) {	// S1, S2, S3 or T
		if(AboveLine(TASLine[3])) {
			strcpy(fieldSymbol, "T");
			if(100 * norm.named.Q / (norm.named.Q + norm.named.an + norm.named.ab + norm.named.orth) > 20.0) {
				strcpy(rootName, "Trachydacite");
			} else {
				strcpy(rootName, "Trachyte");
				if(IsPeralkaline(ox)) {
					IsComenditic(optName, ox);	
				}
			}
		} else if(AboveLine(TASLine[4])) {
			strcpy(fieldSymbol, "S3");
			strcpy(rootName, "Trachyandesite");
			if(IsSodic(ox)) {
				strcpy(optName, "Benmoreite");
			} else {
				strcpy(optName, "Latite");			
			}
		} else if(AboveLine(TASLine[5])) {
			strcpy(fieldSymbol, "S2");
			strcpy(rootName, "Basaltic trachyandesite");
			if(IsSodic(ox)) {
				strcpy(optName, "Mugearite");
			} else {
				strcpy(optName, "Shoshonite");			
			}
		} else {
			strcpy(fieldSymbol, "S1");
			strcpy(rootName, "Trachybasalt");		
			if(IsSodic(ox)) {
				strcpy(optName, "Hawaiite");
			} else {
				strcpy(optName, "Potassic trachybasalt");			
			}
		}
		return;
	}	

	if(AboveLine(TASLine[7])) {	// Ph	
		strcpy(fieldSymbol, "Ph");
		strcpy(rootName, "Phonolite");
		if(IsPeralkaline(ox)) {
			strcpy(optName, "Peralkaline");
		}
		return;
	}

	if(AboveLine(TASLine[9]) && !AboveLine(TASLine[6])) {	// U2, U3
		if(AboveLine(TASLine[8])) {
			strcpy(fieldSymbol, "U3");
			strcpy(rootName, "Tephriphonolite");
		} else if(AboveLine(TASLine[9])) {
			strcpy(fieldSymbol, "U2");
			strcpy(rootName, "Phonotephrite");
		}
		return;
	}

	if (oxide.named.SiO2 > 41.0 && oxide.named.TAlk <= 3.0){	// Pc
		strcpy(fieldSymbol, "Pc");
		strcpy(rootName, "PicroBasalt");
		return;
	}
// Now either F or U1 so check nephelinite and melanephelinite
	
	if(norm.named.ne > 20.0) {
		strcpy(rootName, "Nephelinite");
	} else if (norm.named.ne <= 20.0 && norm.named.ab > 0.0 && norm.named.ab <= 5.0) {
		strcpy(rootName, "Melanephelinite");
	}

	if(oxide.named.SiO2 <= 41.0 || AboveLine(TASLine[6])) {
		strcpy(fieldSymbol, "F");
		if(rootName[0] > 0) return;
		if(norm.named.cs > 0.0) {
			if(norm.named.cs > 10.0 && oxide.named.K2O < oxide.named.Na2O) {
				strcpy(rootName, "Melilitite or olivine melilitite if no kalsilite, otherwise foidite");
			} else if(norm.named.cs > 10.0 && oxide.named.K2O > oxide.named.Na2O && oxide.named.K2O > 2.0) {
				strcpy(rootName, "Potassic melilitite or potassic olivine melilitite");
			} else if (norm.named.cs <= 10.0) {
				strcpy(rootName, "Melilite nephelinite or melilite leucitite");
			} else {
				strcpy(rootName, "Foidite");			
			}
		} else {
			strcpy(rootName, "Foidite");
		}
	} else {
		strcpy(fieldSymbol, "U1");
		if(rootName[0] > 0) return;
			if(norm.named.ol > 10.0)
				strcpy(rootName, "Basanite");
			else
				strcpy(rootName, "Tephrite");
	}
}

// 
//		 IsPeralkaline		(public)
// 
bool
IUGSTAS::IsPeralkaline(SOxidepc &ox)
{
	return ((ox.named.Na2O / molWt_Na2O) + (ox.named.K2O / molWt_K2O)) >
			(ox.named.Al2O3 / molWt_Al2O3);
}

// 
//		 IsComenditic		(public)
// 
bool
IUGSTAS::IsComenditic(char * str, SOxidepc &ox)
{
	if(ox.named.Al2O3 > 1.33 * ox.named.FeO + 4.4) {
		strcpy(str, "Comenditic");
		return true;
	} else {
		strcpy(str, "Pantelleritic");
		return false;
	}
}

// 
//		 IsSodic		(public)
// 
bool
IUGSTAS::IsSodic(SOxidepc &ox)
{
	return ox.named.Na2O - 2.0 >= ox.named.K2O;
}

// 
//		 GetKState		(public)
// 
void
IUGSTAS::GetKState(char * str, SOxidepc &ox)
{
	SLine	KLine[2] =  {{0.085000, -2.880000},		// Line from (48, 1.2) to (68, 2.9) 
						  {0.045000, -1.860000}};	// Line from (48, 0.3) to (68, 1.2) 

	if(AboveLine(ox.named.SiO2, ox.named.K2O, KLine[0])) {
		strcpy(str, "high-K");
	} else if(AboveLine(ox.named.SiO2, ox.named.K2O, KLine[1])) {
		strcpy(str, "medium-K");
	} else {
		strcpy(str, "low-K");
	}
}

// 
//		 AboveLine		(protected)
// 
bool
IUGSTAS::AboveLine(double x, double y, SLine theLine)
{
	return (y > theLine.m * x + theLine.c);
}

// 
//		 AboveLine		(protected)
// 
bool
IUGSTAS::AboveLine(SLine theLine)
{
	return (oxide.named.TAlk > theLine.m * oxide.named.SiO2 + theLine.c);
}

#pragma mark -
// 
//		 Recalculate		(public)
// 
void
IUGSTAS::Recalculate(ERecalculate recalcMode)
{
	Recalculate(oxide, recalcMode); 
}

// 
//		 Recalculate		(public)
// 
void
IUGSTAS::Recalculate(SOxidepc &ox, ERecalculate recalcMode)
{
	if (recalcMode == kDontRecalculate) return;
	
	if(recalcMode == kWaterCO2Free) 
		ox.named.H2OPlus = ox.named.H2OMinus = ox.named.CO2 = 0.0;
	
	double total = GetAnalysisTotal();
	
	if(total > 0.1) {
		total = 100.0/total;
	} else {
		total = 0.0;
	}
	
	ox.named.Total = 0.0;	

	for(int i = 0; i < numOxides - 3; i++ )  {
		ox.wt[i] *= total;
	}
}

// 
//		 GetAnalysisTotal		(public)
// 
double
IUGSTAS::GetAnalysisTotal()
{
	double total = 0.0;
	
	for(int i = 0; i < numOxides - 3; i++ ) total += oxide.wt[i];
	return total;
}

#pragma mark -
// 
//		 CalcCIPW(int usage)		(public)
// 
bool 
IUGSTAS::CalcCIPW(ECO2Usage CO2usage)
{
	return CalcCIPW(CO2usage, oxide, norm);
}

// 
//		 CalcCIPW(int usage)		(public)
// 
bool 
IUGSTAS::CalcCIPW(ECO2Usage CO2usage, SOxidepc &ox, SNormpc &nm)
{
	double	SiO2,	TiO2,	Al2O3,	Fe2O3,		FeO,	MnO,	MgO,	CaO, 
			Na2O,	K2O,	P2O5,	H2OPlus,	CO2,	Zr2O,	Cr2O3,	V2O3, 
			NiO,	CoO,	BaO,	SrO,		Rb2O,	F,		Cl,		S,
			corFac[13],		SiReq,	SiDef,		MgRatio;
	int	i;

	ox.named.Total = 0.0;
	for(i = 0; i < numOxides - 3; i++ ) {
		ox.named.Total += ox.wt[i];
	}	
	
	ox.named.TAlk = ox.named.Na2O + ox.named.K2O;

// Calulates Molecular proportions 
	SiO2 = ox.named.SiO2 / molWt_SiO2;		TiO2 = ox.named.TiO2 / molWt_TiO2;	
	Al2O3 = ox.named.Al2O3 / molWt_Al2O3;	Fe2O3 = ox.named.Fe2O3 / molWt_Fe2O3;	
	FeO = ox.named.FeO / molWt_FeO;			MnO = ox.named.MnO / molWt_MnO;
	MgO =ox.named.MgO / molWt_MgO;			CaO = ox.named.CaO / molWt_CaO;	
	Na2O = ox.named.Na2O / molWt_Na2O;		K2O = ox.named.K2O / molWt_K2O;
	P2O5 = ox.named.P2O5 / molWt_P2O5;		H2OPlus = ox.named.H2OPlus / molWt_H2OPlus;
    CO2 = ox.named.CO2 / molWt_CO2;			Zr2O = ox.named.ZrO2 / molWt_ZrO2;	
   	Cr2O3 = ox.named.Cr2O3 / molWt_Cr2O3;	V2O3 = ox.named.V2O3 / molWt_V2O3;
   	NiO = ox.named.NiO / molWt_NiO;			CoO = ox.named.CoO / molWt_CoO;
   	BaO = ox.named.BaO / molWt_BaO;			SrO = ox.named.SrO / molWt_SrO;
   	Rb2O = ox.named.Rb2O / molWt_Rb2O;		F = ox.named.F / molWt_F;
   	Cl = ox.named.Cl / molWt_Cl;			S = ox.named.S / molWt_S;

//  Summation of substituting oxides and molecular weight corrections 
	K2O = K2O + Rb2O;
	CaO = CaO + SrO + BaO;
	Fe2O3 = Fe2O3 + V2O3;
	FeO = FeO + MnO + NiO + CoO;
	if (K2O != 0.0) corFac[2] = Rb_K_CorFac * Rb2O / K2O; 
		else corFac[2] = 0.0;
	if (CaO != 0.0) corFac[3] = (Sr_Ca_CorFac * SrO + Ba_Ca_CorFac * BaO) / CaO;
		else corFac[3] = 0.0;
	if (Fe2O3 != 0.0) corFac[4] = V_Fe3_CorFac * V2O3 / Fe2O3;
		else corFac[4] = 0.0;
	if (FeO != 0.0) corFac[5] = (Co_Fe2_CorFac * CoO + Ni_Fe2_CorFac * NiO + Mn_Fe2_CorFac * MnO) / FeO;
		else corFac[5] = 0.0;
	
	nm.named.Other = ox.named.Other;

//  Calculation of cipw norm by the method of Kelsey, Min. Mag. 1965, p.276.
	nm.named.Z = Zr2O;
	nm.named.ap = Minimum(Minimum(0.2 * CaO, P2O5 / 3.0), F);
		CaO -= 5.0 * nm.named.ap;
		P2O5 -= 3 * nm.named.ap;
		if(P2O5 <= kVerySmall) P2O5 = 0.0;
		F -= nm.named.ap;
	nm.named.fr = Minimum(0.5 * F, CaO);
		CaO -= nm.named.fr;
		F -= (2 * nm.named.fr);
	nm.named.hap = Minimum(0.2 * CaO, P2O5 / 3.0);		
		CaO -= 5.0 * nm.named.hap;
		P2O5 -= 3 * nm.named.hap;
		if(P2O5 <= kVerySmall) P2O5 = 0.0;
		H2OPlus -= nm.named.hap; 
	nm.named.hl = Minimum(Na2O, Cl);
		Na2O -= nm.named.hl;
		Cl -= nm.named.hl;
	nm.named.pr = Minimum(0.5 * S, FeO);
		FeO -= nm.named.pr;
		S -= (2 * nm.named.pr);
	 
// Deal with CO2
	switch(CO2usage)	{		
		case kCO2AsCaMgFeCarb:					// Calculate as CaCO3, MgCO3, FeCO3 
			nm.named.cc = Minimum(CaO, CO2);
				CaO -= nm.named.cc;
				CO2 -= nm.named.cc;
			nm.named.mag = Minimum(MgO, CO2);
				MgO -= nm.named.mag;
				CO2 -= nm.named.mag;
			nm.named.sid = Minimum(FeO, CO2);
				FeO -= nm.named.sid;
				CO2 -= nm.named.sid;
			break;
	   	
		case kCO2AsOthers:
			nm.named.Other = ox.named.Other + CO2 * molWt_CO2;	// Add CO2 to others
			CO2 = 0.0;
			break;
	   	
		case kCO2AsNaCarb:
			nm.named.nc = Minimum(0.5 * Na2O, CO2);				// Calculate as Na2CO3
				Na2O -= 2 * nm.named.nc;
				CO2 -= nm.named.nc;
			if(CO2 > kVerySmall)	{							// CO2 remaining so calculate as
				nm.named.cc = Minimum(CaO, CO2);				// CaCO3, MgCO3, FeCO3 
					CaO -= nm.named.cc;
					CO2 -= nm.named.cc;
				nm.named.mag = Minimum(MgO, CO2);
					MgO -= nm.named.mag;
					CO2 -= nm.named.mag;
				nm.named.sid = Minimum(FeO, CO2);
					FeO -= nm.named.sid;
					CO2 -= nm.named.sid;
			}
			break;
	} 
// Allot other normative constituents 
	nm.named.cm = Minimum(FeO, 0.5 * Cr2O3);
		FeO -= nm.named.cm;
		Cr2O3 -= 2 * nm.named.cm;
	nm.named.il = Minimum(FeO, TiO2);
		FeO -= nm.named.il;
		TiO2 -= nm.named.il;
	nm.named.orth = Minimum(K2O, Al2O3);
		Al2O3 -= nm.named.orth;
		nm.named.ks = 0.5 * (K2O - nm.named.orth);
	nm.named.ab = Minimum(Na2O, Al2O3);
		Al2O3 -= nm.named.ab;
		Na2O -= nm.named.ab;
	nm.named.ac = Minimum(Fe2O3, Na2O);
		Fe2O3 -= nm.named.ac;
		nm.named.ns = 0.5 * (Na2O - nm.named.ac);
	nm.named.an = Minimum(0.5 * Al2O3, CaO);
		nm.named.C = Al2O3 - 2 * nm.named.an;
		CaO -= nm.named.an;
	nm.named.tn = Minimum(TiO2, CaO);
		CaO -= nm.named.tn;
		nm.named.ru = TiO2 - nm.named.tn;
	nm.named.mt = Minimum(FeO, 0.5 * Fe2O3);
		FeO -= nm.named.mt;
		nm.named.hm = Fe2O3 - 2 * nm.named.mt;
	MgRatio = FeO + MgO;
	nm.named.di = Minimum(MgRatio, CaO);
		nm.named.wo = CaO - nm.named.di;
		nm.named.hy = MgRatio - nm.named.di;
	
//  Find amount of SiO2 required 
	SiReq = nm.named.orth + nm.named.ab + nm.named.ns + nm.named.ks + nm.named.tn 
			+ nm.named.wo + nm.named.Z + nm.named.hy 
			+ 2.0 * (nm.named.orth + nm.named.ab + nm.named.an + nm.named.ac + nm.named.di); 
	if(SiO2 > SiReq) {			// Oversaturated 
		nm.named.Q = SiO2 - SiReq;
		nm.named.SiUnd = 0.0;					
	} else	{							// uUndersaturated 
		nm.named.Q = 0.0;
		SiDef = SiReq - SiO2;
		nm.named.SiUnd = SiDef * molWt_fsd;	// Set Si Undersaturation
		SiReq = 0.5 * nm.named.hy;
		nm.named.ol = Minimum(SiReq,SiDef);	// Desilicate minerals
			nm.named.hy -= 2 * nm.named.ol;
			SiDef -= nm.named.ol;
		if(SiDef > kVerySmall)	{
			nm.named.pf = Minimum(SiDef,nm.named.tn);
				nm.named.tn -= nm.named.pf;
				SiDef -= nm.named.pf;
			if(SiDef > kVerySmall)	{
				nm.named.ne = 0.5 * SiDef;
				if(SiDef > 2.0 * nm.named.ab)nm.named.ne = nm.named.ab;
				nm.named.ab = nm.named.ab - nm.named.ne;
				SiDef = SiDef - 2.0 * nm.named.ne;
				if(SiDef > kVerySmall)	{
					nm.named.lc = Minimum(SiDef, nm.named.orth);
						nm.named.orth -= nm.named.lc;
						SiDef -= nm.named.lc;
					if(SiDef > kVerySmall)	{
						nm.named.cs = Minimum(SiDef, 0.5 * nm.named.wo);
							nm.named.wo -= 2 * nm.named.cs;
							SiDef -= nm.named.cs;
						if(SiDef > kVerySmall)	{
							SiReq = 0.5 * Minimum(SiDef, nm.named.di);
								nm.named.cs += SiReq;
								nm.named.ol += SiReq;
								nm.named.di -= 2 * SiReq;
								SiDef -= 2 * SiReq;
							if(SiDef > kVerySmall)	{
								nm.named.kp = Minimum(SiDef, nm.named.lc);
									nm.named.lc -= nm.named.kp;
									SiDef -= nm.named.kp;
								if(SiDef > kVerySmall) nm.named.Fsd = - SiDef;
							}
						}
					}
				}
			}
		}
	}
// Molecular weight corrections and conversion to weight percents 
	corFac[6] = corFac[3] + corFac[3];
	corFac[7] = 2.0 * corFac[4] + corFac[5];
	corFac[8] = 5.0 * corFac[3];
	if(MgRatio > kVerySmall) {
		MgRatio = FeO / MgRatio;
		corFac[10] = (Fe2_Mg_CorFac + corFac[5]) * MgRatio;
		nm.named.MgRatio = 100.0 * (1.0 - MgRatio);		// Mg/Mg+Fe
	} else {
		nm.named.MgRatio = 0.0;		// Mg/Mg+Fe
		corFac[10] = 0.0;
	}
	corFac[9] = corFac[10] + corFac[3];
	corFac[11] = corFac[10] + corFac[10];
	corFac[12] = corFac[2] + corFac[2];

// Form wt% Normative components 
	nm.named.Q *= molWt_Q;
	nm.named.C *= molWt_C;
	nm.named.Z *= molWt_Z;
	nm.named.orth *= (molWt_or + corFac[2]);
	nm.named.ab *= molWt_ab;
	nm.named.an *= (molWt_an + corFac[3]);
	nm.named.ne *= molWt_ne;
	nm.named.lc *= (molWt_lc + corFac[2]);
	nm.named.kp *= (molWt_kp + corFac[2]);
	nm.named.hl *= molWt_hl;
	nm.named.nc *= molWt_nc;
	nm.named.ac *= (molWt_ac + corFac[4]);
	nm.named.ns *= molWt_ns;
	nm.named.ks *= (molWt_ks + corFac[12]);
	nm.named.di *= (molWt_di + corFac[9]);
	nm.named.hy *= (molWt_hy + corFac[10]);
	nm.named.wo *= (molWt_wo + corFac[3]);
	nm.named.ol *= (molWt_ol + corFac[11]);
	nm.named.cs *= (molWt_cs + corFac[6]);
	nm.named.cm *= (molWt_cm + corFac[5]);
	nm.named.hm *= (molWt_hm + corFac[4]);
	nm.named.mt *= (molWt_mt + corFac[7]);
	nm.named.il *= (molWt_il + corFac[5]);
	nm.named.tn *= (molWt_tn + corFac[3]);
	nm.named.pf *= (molWt_pf + corFac[3]);
	nm.named.ru *= molWt_ru;
	nm.named.ap *= (molWt_ap + corFac[8]);
	nm.named.hap *= (molWt_hap + corFac[8]);
	nm.named.fr *= (molWt_fl + corFac[3]);
	nm.named.pr *= (molWt_py + corFac[5]);
	nm.named.cc *= (molWt_cc + corFac[3]);
	nm.named.mag *= molWt_mag;
	nm.named.sid *= (molWt_sid + corFac[5]);
	nm.named.H2OPlus = H2OPlus * molWt_H2OPlus;
	nm.named.H2OMinus = ox.named.H2OMinus;  
	nm.named.Fsd *= molWt_fsd;
	nm.named.exP2O5 = P2O5 * molWt_P2O5;	
	nm.named.exCO2 = CO2 * molWt_CO2;	
	nm.named.exCr2O3 = Cr2O3 * molWt_Cr2O3;	
	nm.named.exF = F * molWt_F;	
	nm.named.exCl = Cl * molWt_Cl;	
	nm.named.exS = S * molWt_S;

// Sum normative values
	for (i = 0; i < numNormValues - 3; i++) nm.named.Total += norm.wt[i];
	
//  Correct original total for F, S and Cl taking any excess into account 
	ox.named.OxEq = - (ox.named.F - nm.named.exF) * oxyEquiv_F 
								- (ox.named.Cl - nm.named.exCl) * oxyEquiv_Cl 
								- (ox.named.S - nm.named.exS) * oxyEquiv_S;

// Check two totals for errors
	if(fabs(ox.named.Total + ox.named.OxEq - nm.named.Total) > 0.001) {
		cout << bellChar;
		return false;		// Unexpected error with total
	}
	
// 	A fudge because of problems with CodeWarrior being unable to stream certain values
	for (i = 0; i < numOxides; i++) {Fudge(oxide.wt[i]);}			
	for (i = 0; i < numNormValues; i++) {Fudge(norm.wt[i]);}			

	return true;		// Total OK

}

// 
//		 Fudge		(protected)
// 
void 
IUGSTAS::Fudge(double &inValue)
{
	// fudge to deal with streaming and conversion problem in CodeWarrior Library
	int val;
	
	if(inValue > 0.0) {			
		val = (int)(inValue * 100.0 + 0.5);
		inValue = double(val) / 100.0; 
	}
}

#pragma mark -

// 
//		 CheckOutputFile		(protected)
// 
void
IUGSTAS::CheckOutputFile(char *fName, int &nos, char *routineName)
{
	char temp[200];

	ostrstream ii(temp, 200);
	ii << routineName << "(";
	if(fName[0] > 0) ii << "\"" << fName << "\", ";
	ii << nos << ')' << '\0';
	
	if(nos < 1 || nos > numOutputFiles) {
		AbortProgramingError(temp, "Output file numbers must be in range 1 to ", 
														numOutputFiles, ".");
	}
	
	nos--;
	
	if(fName[0] > 0) {
		if(outputDefined[nos]) {
			AbortProgramingError(temp, "Output file number ", nos + 1, " has already been opened.");
		} else if(FileInUse(fName)) {
			AbortProgramingError(temp, "The named file is already in use.", "");		
		}
	} else if(strncmp("Close", routineName, 5) != 0 && !outputDefined[nos]) {
		AbortProgramingError(temp, "Output file number ", nos + 1, " is not currently open.");
	};
}

// 
//		 FileInUse		(protected)
// 
bool
IUGSTAS::FileInUse(char* fName)
{
	int		i;
	
	if(strcmp(fName, (char *)&inputFileName) == 0) return true;
	
	for (i = 0; i < numOutputFiles; i++) {
		if(strcmp(fName, (char *)&outputFileName[i]) == 0) return true;
	}
	return false;
}

// 
//		 AbortDataError		(protected)
// 
void
IUGSTAS::AbortDataError(char* inStr, char* mess)
{
	cout << endl << endl << endl << sixAst << "DATA ERROR with string \"" << inStr << "\"";
	cout << " in row " << (inputRow + 1) << ", column " << (inputCol + 1) << ".";
	cout << endl << endl << sixAst << mess;
	Abort();
}

// 
//		 AbortProgramingError		(protected)
// 
void
IUGSTAS::AbortProgramingError(char* routine, char* mess1, char* mess2)
{
	AbortProgramingError(routine, mess1);
	if(mess2[0]) cout << endl << endl << sixAst << mess2;
	Abort();
}

// 
//		 AbortProgramingError		(protected)
// 
void
IUGSTAS::AbortProgramingError(char* routine, char* mess1, int n, char* mess2)
{
	AbortProgramingError(routine, mess1);
	cout << n << mess2;
	Abort();
}

// 
//		 AbortProgramingError		(protected)
// 
void
IUGSTAS::AbortProgramingError(char* routine, char* mess)
{
	cout << endl << endl << endl << sixAst << "PROGRAMING ERROR in call to \"";
	cout << routine << "\".";
	cout << endl << endl << sixAst << mess;
}

// 
//		 UnexpectedWriteError		(protected)
// 
void
IUGSTAS::UnexpectedWriteError(char* routine, char* analName)
{
	cout << endl << bellChar << sixAst<< "UNEXPECTED WRITE ERROR in routine \"";
	cout << routine << "\" - with analysis:" << endl << warnSpaces << analName << endl;
}

// 
//		 Abort		(protected)
// 
void
IUGSTAS::Abort()
{
	cout << endl << endl << sixAst << "Execution aborted." << endl;
	exit(1);
}


#pragma mark -

// 
//		 IsPositiveNumber
// 
bool 
IUGSTAS::IsPositiveNumber(char *inStr)
{
	if(inStr[0] == '\0') return false;
	
	int numDP = 0;
	
	for (int i = 0; i < strlen(inStr); i++) {
		char ch = inStr[i];
		if(ch == '.') {
			numDP++;
		} else if(!isdigit(ch)) {
			return false;
		}
	}
	return numDP < 2;
}

// 
//		 Warning	
// 
void 
IUGSTAS::Warning(char* mess1, char* mess2, char* mess3, char* mess4)
{
	cout << bellChar << sixAst << warnMess;
	cout << mess1 << mess2 << mess3;
	if(strlen(mess4) > 0) {
		cout << endl << sixAst << warnSpaces << mess4; 
	}
	cout << endl << endl;
}

